import os
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
import binascii

# AES ECB mode without IV

key = os.environ.get('CIPHER_PRIVATE_KEY')  # Must Be 16 char for AES128


def encrypt(raw):
    raw = pad(raw.encode(), 16)
    cipher = AES.new(key.encode('utf-8'), AES.MODE_ECB)
    encrypted_data = cipher.encrypt(raw)
    encrypted_data_hex = binascii.hexlify(encrypted_data).decode('utf-8')
    return encrypted_data_hex


def decrypt(enc):
    try:
        enc = binascii.unhexlify(enc.encode('utf-8'))
        cipher = AES.new(key.encode('utf-8'), AES.MODE_ECB)
        decrypted = unpad(cipher.decrypt(enc), 16)
        return decrypted.decode('utf-8', 'ignore')
    except (ValueError, TypeError):
        # Handle decryption errors, such as invalid input or incorrect padding
        return None



import json
import traceback
from datetime import datetime
from flask import request


class UserObject:
    """
    Custom user class
    Helps to convert dict to class object instance.
    """

    is_authenticated = False
    is_staff = False
    is_superuser = False
    is_agent = False
    permissions = list()

    def __init__(self, dict1):
        self.__dict__.update(dict1)


def get_user_obj(data):
    """
    using 'json.loads' method and passing 'json.dumps'
    method and custom object hook as arguments
    """
    if not isinstance(data, dict):
        raise TypeError("get_user_obj 'data' argument must be a dict.")
    return json.loads(json.dumps(data), object_hook=UserObject)


def save_error_es_db():
    from . import EsClient
    just_the_string = traceback.format_exc()

    body = {
        "Path": request.path,
        "Method": request.method,
        "Entity": "orders",
        "Data": str(just_the_string),
        "StatusCode": 500,
        "Timestamp": datetime.now()
    }
    EsClient.index(index='error_logs', body=body)
    return True



from __future__ import annotations
from elasticsearch import Elasticsearch
from settings import *

EsClient = Elasticsearch(ELASTICSEARCH_URL)


def register_events():
    from . import events
    activity_events = events.ActivityEvent()
    activity_events.initialise_events()


register_events()



"""
This filter file contains custom Filter classes, SQLAlchemy model classes and SQLAlchemy query filter.
note: File is created for code optimization purpose.
"""
from flask import request
from flask_sqlalchemy import BaseQuery


class Filter(object):
    """
    Flask filter class,
    Inherit this class to apply filters on queryset.

    e.g.

    class ModelFilter(Filter):
        model = model_name
        def add_filter(self, *args, **kwargs):
            queryset = self.get_queryset()
            queryset = queryset.filter_by(**self.request_params)
            return queryset

    use this filter class in your API view or function.

    class APIView():
        filter_class = ModelFilter
        def get():
            queryset = self.filter_class(queryset).apply()

    """

    model = None
    query_class = BaseQuery

    def __init__(self, queryset):
        self.filter_added = False
        self.result = None
        self.request_params = request.args.copy()

        if not self.model:
            raise TypeError("<class '%s'> requires model name." % self.__class__.__name__)

        if not isinstance(queryset, self.query_class):
            raise TypeError("queryset must be instance of <class '%s'>" % self.query_class.__name__)

        self.queryset = queryset

    def get_queryset(self):
        return self.queryset

    def all(self):
        """
        Return the results as a list object.
        """
        if not self.filter_added:
            raise TypeError("Queryset filter not applied on <class '%s'>, "
                            "Please '.apply()' filter first." % self.__class__.__name__)
        return self.result.all()

    def add_filter(self, *args, **kwargs):
        """
        Override this func to add custom filters on queryset.
        """
        if not self.filter_added:
            raise TypeError("Queryset filter not applied on <class '%s'>, "
                            "Please '.apply()' filter first." % self.__class__.__name__)
        queryset = self.get_queryset()
        return queryset.filter(*args, **kwargs)

    def apply(self, *args, **kwargs):
        """
        Apply filter on queryset and return results.
        """
        self.filter_added = True
        self.result = self.add_filter(*args, **kwargs)
        return self.result



"""
Contains:
SqlAlchemy model events before_create, after_create, before_update, after_update.
"""
import importlib
import json
import sys
from decimal import Decimal
import enum
import inspect as module_inspect
from datetime import datetime, date

from elasticsearch_dsl import Search
from flask import request
from flask_sqlalchemy.model import Model
from sqlalchemy import event as alchemy_event, inspect
from . import EsClient, exceptions, ACTIVITY_INDEX, EXCLUDE_MODEL_EVENTS, EVENTS, APP_MODELS


class BaseEvents(object):
    """
    Base events class.
    """
    events = []

    def initialise_events(self):
        if not self.events:
            raise exceptions.NotImplementedException("Please register models events through EventRegistry")

        for event in self.events:
            alchemy_event.listen(event['model'], event['name'], getattr(self, event['name']))

    def before_update(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")

    def after_update(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")

    def after_insert(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")

    def before_insert(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")

    def before_delete(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")

    def after_delete(self, mapper, connection, target):
        raise exceptions.NotImplementedException("Not implimented")


class EventRegistry(BaseEvents):

    @classmethod
    def register(cls, model, events: list):
        for event in events:
            cls.events.append({'model': model, 'name': event})


class ActivityMixin:
    """
    Activity event class,
    register event class in __init__.py file

    e.g.
    from . import events
    activity_events = events.ActivityEvent()
    activity_events.initialise_events()
    """

    @classmethod
    def save_activity(cls, table_name, activity, new_data, old_data=None, **kwargs):
        """
        Push user activity logs to Elasticsearch.
        """

        if old_data is None:
            old_data = {}

        request_id = request.headers.get("Request-Id")
        activity_data = {
            "Table": table_name,
            "OldData": old_data,
            "NewData": new_data,
            "Activity": activity
        }
        search = Search(using=EsClient, index=ACTIVITY_INDEX).query(
            "match", RequestID=request_id)
        try:
            response = search.execute()
            hits = response.hits
            if hits:
                # Retrieve the first matching document and update the activity array
                hit = hits[0]
                index_id = hit.meta.id
                existing_activities = json.loads(getattr(hit, 'Activities', "[]"))
                existing_activities.append(activity_data)
                update_body = {
                    "doc": {
                        "Activities": json.dumps(existing_activities)
                    }
                }
                EsClient.update(index=ACTIVITY_INDEX, id=index_id, body=update_body)
                EsClient.indices.refresh(index=ACTIVITY_INDEX)
            else:
                log_body = {
                    "RequestID": request_id,
                    "Activities": json.dumps([activity_data])
                }
                # Create a new document with the request ID and the activity array
                EsClient.index(index=ACTIVITY_INDEX, body=log_body)
                EsClient.indices.refresh(index=ACTIVITY_INDEX)
        except Exception as error:
            print("Error saving activity: %s" % str(error))

    @classmethod
    def get_history(cls, mapped_obj, target, column):
        """
        This function helps to get history/old value of a column
        if any change available.
        SqlAlchemy model object hold the old value until changed in DB.
        """
        is_changed = False
        old_value = None
        column_attr = getattr(mapped_obj.attrs, column.key)
        history = column_attr.history
        latest_value = getattr(target, column.key)

        # If a scalar attribute has been changed, the old value will be in
        # deleted.
        if history.deleted:
            is_changed = True
            old_value = history.deleted[-1]
        if history.added:
            is_changed = True

        return {'is_changed': is_changed, 'latest': latest_value, 'old': old_value}

    @classmethod
    def get_value(cls, value):
        """
        This function is used to change type of value if value is non JSON serializable.
        Update this function if you need to change value type.
        """

        try:
            # Attempt to serialize the value using json.dumps()
            return json.loads(json.dumps(value))
        except (TypeError, ValueError):
            # Handle non-serializable values
            if isinstance(value, enum.Enum):
                return value.value if value else None
            elif isinstance(value, (datetime, date)):
                return value.isoformat()
            elif isinstance(value, Decimal):
                return float(value)
            else:
                return None

    def get_model_data(self, target, history=False):
        """
        Get model/columns data from model instance.
        """
        old_data = {}
        new_data = {}

        # SqlAlchemy inspect will provide mapper obj
        # With mapper object we can get model attributes and columns
        mapper_obj = inspect(target)
        for column in mapper_obj.mapper.column_attrs:
            history_data = self.get_history(mapper_obj, target, column)
            if history_data.get('is_changed'):
                if history:
                    old_data[column.key] = self.get_value(history_data.get('old'))
                new_data[column.key] = self.get_value(history_data.get('latest'))

        return new_data, old_data


class ActivityEvent(BaseEvents, ActivityMixin):
    """
    Contains default events to save activity,
    Inherit this activity class to add more events.
    """

    def before_update(self, mapper, connection, target):
        """
        Receiver function before update.
        """
        activity = "Update"
        new, old = self.get_model_data(target, history=True)
        self.save_activity(
            table_name=target.__tablename__,
            activity=activity,
            new_data=new,
            old_data=old
        )

        return target

    def after_insert(self, mapper, connection, target):
        """
        Receiver function after insert,
        Push user activity logs to Elasticsearch.
        """
        activity = "Create"
        new, old = self.get_model_data(target)
        self.save_activity(
            table_name=target.__tablename__,
            activity=activity,
            new_data=new
        )

        return target


def register_models():
    """
    Auto Register model to activity event.
    Check settings file to exclude models from activity events.
    """
    model_events = EventRegistry()
    module = importlib.import_module(APP_MODELS)
    for name, obj in module_inspect.getmembers(sys.modules[module.__name__]):
        if module_inspect.isclass(obj) and issubclass(obj, Model) and name not in EXCLUDE_MODEL_EVENTS:
            model_events.register(model=obj, events=EVENTS)


register_models()



import os
from functools import wraps
from flask import request, jsonify
from flask_babel import lazy_gettext as _
from .authentication import get_authorization_header, authenticate, get_internal_key_header
from .encryption import decrypt
from .exceptions import AuthException, APPException


INTERNAL_SERVICE_KEY = os.environ.get('INTERNAL_SERVICE_KEY')


def is_authenticated(func):
    """
    Decorator for authenticate request,
    If token is authenticated user details added into "request.user".
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        headers = request.headers
        token = get_authorization_header(headers)
        if token:
            try:
                user = authenticate(token)
                setattr(request, 'user', user)
            except AuthException as error:
                return error.as_json()
        else:
            return AuthException().as_json()
        return func(*args, **kwargs)

    return wrapper


def is_agent(func):
    """
    Decorator for authenticate agent,
    If token is authenticated agent set to True into "request.user.is_agent".
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not bool(getattr(request, 'user', None) and getattr(request.user, 'is_agent', None)):
            return jsonify({"message": _("Permission denied.")}), 401
        return func(*args, **kwargs)

    return wrapper


def admin_authenticated(token):
    response = {"is_admin": False, "error": None}
    try:
        try:
            token = decrypt(token)
        except APPException:
            response['error'] = "Invalid Token."
            return response
        token = "Bearer %s" % token
        user, permissions = authenticate(token)
        if user.user_type == "ADMIN":
            response['is_admin'] = True
    except Exception as error:
        response['error'] = str(error)
    return response


def authenticated_request(allow_internal_request=False):
    """
    Decorator function to authorize the API request,
    this decorator authorize internal request and token request.
    """

    def internal_request_deco(func):
        try:
            headers = getattr(request, 'headers', {})
        except:
            headers = dict()

        request_internal_key = get_internal_key_header(headers)

        if allow_internal_request and request_internal_key:
            @wraps(func)
            def wrapper(*args, **kwargs):
                if request_internal_key == INTERNAL_SERVICE_KEY:
                    setattr(request, 'is_internal_request', True)
                else:
                    return is_authenticated(func)
                return func(*args, **kwargs)
            return wrapper
        else:
            return is_authenticated(func)

    return internal_request_deco



"""
Contains override Exception class.
"""
from flask import jsonify
from flask_babel import lazy_gettext as _
from . import app, utils
from .logger import logger
from .response import CustomResponse


class APPException(Exception):
    """
    Provide `.status_code` and `.default_detail` properties.
    """
    code = 500
    default_detail = _('An error occurred.')
    error_field = 'error'

    def __init__(self, detail=None, status_code=None):
        self.detail = detail or self.default_detail
        self.messages = [self.detail] if isinstance(self.detail, (str, bytes)) else self.detail
        self.status_code = status_code or self.code
        super(APPException, self).__init__(detail)

    def __str__(self):
        """
        Return exception str error.
        """
        return str(self.detail)

    def get_exception_message(self, args: Exception.args):
        """
        Return Exception messages as dict.
        :param args: `Exception.args`
        :return: dict
        """
        all_exception_detail = dict()
        if isinstance(args, tuple):
            for arg in args:
                if getattr(arg, 'args', None):
                    all_exception_detail.update(self.get_exception_message(arg.args))
                elif isinstance(arg, dict):
                    all_exception_detail.update(arg)
                else:
                    all_exception_detail.update({self.error_field: str(arg)})
        return all_exception_detail

    @property
    def messages_dict(self):
        """
        Return `exceptions` of type Dict.
        """
        exception_detail = self.get_exception_message(self.args)

        return exception_detail

    def as_json(self):
        """
        Return `flask.jsonify` response.
        """
        response = jsonify(self.messages_dict)
        response.status_code = self.status_code
        return response


class AuthException(APPException):
    """
    Token Authentication exception.
    """
    code = 401
    default_detail = _('No authentication credentials were provided.')
    error_field = 'message'


class APIException(APPException):
    """
    API exception.
    """
    code = 400
    default_detail = _('Something went wrong.')
    error_field = 'message'


class NotImplementedException(Exception):
    pass


@app.errorhandler(APPException)
def app_exception_handler(error):
    logger.critical('[ERROR]--{}'.format(str(error)))
    return error.as_json()


@app.errorhandler(Exception)
def all_exception_handler(error):
    utils.save_error_es_db()
    logger.critical('[ERROR]--{}'.format(str(error)))
    response = jsonify({'message': _('Something went wrong')})
    response.status_code = 400
    return response


@app.after_request
def after_request(response):
    if response.content_type == 'application/json':
        return CustomResponse(response).render
    return response



import json
import logging


def logging_override(name: str, extra=None):
    if extra is None:
        extra = {}
    c_logger = logging.getLogger(name)
    c_logger.setLevel(logging.DEBUG)

    stream_handler = logging.StreamHandler()

    stream_formatter = {
        "time": "%(asctime)s", "level": "%(levelname)s", "message": "%(message)s"}
    stream_formatter = logging.Formatter(json.dumps(stream_formatter))
    stream_handler.setFormatter(stream_formatter)

    c_logger.addHandler(stream_handler)
    c_logger = logging.LoggerAdapter(c_logger, extra)
    return c_logger


logger = logging_override("json")



"""Services to connect with service."""
import math
from elasticsearch_dsl import Search, Q
from . import EsClient, SERVICE_CACHE_INDEX
from .exceptions import APIException


class ServiceBase(object):
    def __init__(self, service_id, units, raise_exception=False):
        self.service_id = service_id
        self.units = units or 1
        self._total_price = 0.0
        self._total_commission = 0.0
        self._total_discount = 0.0
        self._tax_amount = 0.0
        self._coupon_code = None
        self._service_details = None
        self.is_valid_service = False

        index_query = Search(
            using=EsClient, index=SERVICE_CACHE_INDEX
        ).query("match", id=self.service_id)

        response = index_query.execute()
        if len(response.hits) < 1:
            if raise_exception:
                raise APIException("Service not found", 404)
        else:
            self._service_details = response.hits[0]
            self.is_valid_service = True

    @property
    def total_commission(self):
        return self._total_commission

    @property
    def service_details(self):
        if not self._service_details:
            raise APIException("Service details not found")
        return self._service_details

    @property
    def data(self):
        return self.service_details.to_dict()

    @property
    def total_discount(self):
        return self._total_discount

    @property
    def coupon_code(self):
        return self._coupon_code

    @property
    def tax_details(self):
        return self.service_details.tax.to_dict()

    @property
    def total_price(self):
        return self._total_price


class PriceCatalog(ServiceBase):
    """
    Service price class to fetch service fee structure from `Elasticsearch` index.
    service_id: Respective service id,
    units: Units/quantity of service
    -----------------------------------------------------------------------------
    Call `ServicePrice(service_id, units).get_fee_details()` to get service fee details.

    """
    FLAT = 'FLAT'
    TIER = 'TIER'

    def get_fee_details(self, include_tax=False):
        """
        Return Service fee details based on service fee structure (FLAT/TIER).
        """
        response = dict()
        units = self.units
        obj = self.service_details
        response['service_name'] = obj.name
        if obj.fee_structure == ServicePrice.FLAT:

            if not obj.service_fee_detail:
                raise APIException(
                    'Flat service fee structure pricing not available'
                )

            flat = obj.service_fee_detail[0]
            self._total_price = units * flat.total_cost
            response['hard_cost'] = flat.hard_cost
            response['fees'] = flat.fees
            response['set_up_cost'] = flat.set_up_cost
            response['is_hard_cost_pay_by_dcg'] = flat.is_hard_cost_pay_by_dcg
            response['total_price'] = self._total_price
            response['total_commission'] = self.get_flat_commission(self._total_price)
            response['amount_break_down'] = [flat.to_dict()]

        elif obj.fee_structure == ServicePrice.TIER:
            tiers = obj.service_tier_fee_detail
            tiers = list(filter(lambda k: k.unit_from <= units, tiers))
            tiers = sorted(tiers, key=lambda k: k.unit_from, reverse=True)

            if not tiers:
                raise APIException(
                    'Tier service fee structure pricing not available'
                )

            # Removed as per client requirement
            # if tiers[0].unit_to and tiers[0].unit_to < units:
            #     raise APIException(
            #         'You can take max of ' + str(tiers[0].unit_to) + ' units'
            #     )

            total_price, total_commission = self.tier_cal_generator(tiers)

            response['total_price'] = total_price
            response['total_commission'] = total_commission
            response['amount_break_down'] = [tier.to_dict() for tier in tiers]

        service_tax = obj.tax
        if service_tax.tax_percentage:
            self._tax_amount = self.get_service_tax(self._total_price)
            if include_tax:
                self._total_price += self._tax_amount

        response['total_price'] = self._total_price
        response['tax_amount'] = self._tax_amount
        response['tax_code'] = service_tax.tax_code
        response['tax_percentage'] = service_tax.tax_percentage

        return response

    def get_service_tax(self, total_price):
        service_tax = self.service_details.tax
        tax_amount = self.percentage(total_price, service_tax.tax_percentage)
        return tax_amount

    def tier_cal_generator(self, tiers):
        """
        Return/calculate tier based service price and commission.
        """
        units = self.units
        self._total_price = 0.0
        self._total_commission = 0.0

        for obj in tiers:
            if obj.price_type == "GROUP":
                group_of = units - (obj.unit_from - 1)
                obj.total_price = obj.price * (math.ceil(group_of / obj.group_by))
                self._total_commission += self.get_tier_commission(obj.total_price, obj)
                self._total_price += obj.total_price
                continue

            if obj.price_type == "TOTAL":
                obj.total_price = obj.price
                self._total_commission += self.get_tier_commission(obj.total_price, obj)
                self._total_price += obj.total_price
                continue

            if not obj.unit_to:
                unit = units - (obj.unit_from - 1)
            else:
                if obj.unit_to and (obj.unit_from <= units <= obj.unit_to):
                    unit = units - (obj.unit_from - 1)
                else:
                    unit = obj.unit_to - (obj.unit_from - 1)

            obj.total_price = unit * obj.price
            self._total_commission += self.get_tier_commission(obj.total_price, obj)
            self._total_price += obj.total_price
        return self._total_price, self._total_commission

    @staticmethod
    def percentage(part, whole):
        return (float(part) * float(whole)) / 100

    def get_flat_commission(self, service_price):
        """
        If Service is flat calculate commission from ServiceCommissions.
        If The commission_type is Flat return commission as figure value.
        If The commission_type is PERCENTAGE Calculate commission
        by ((service_total_price*commission.figure_value)/100).
        """
        total_commission = 0.0
        service = self.service_details
        if service.fee_structure == 'FLAT':
            if not service.service_commissions:
                return total_commission

            commission_obj = service.service_commissions[0]

            if commission_obj.commission_type == "FLAT":
                total_commission = commission_obj.figure_value
            else:
                total_commission = self.percentage(service_price, commission_obj.figure_value)

        self._total_commission = total_commission

        return total_commission

    def get_tier_commission(self, service_price, service_tier):
        """
        If Service is tier calculate commission from ServiceTierCommissions.
        If The commission_type is Flat return commission as figure value.
        If The commission_type is PERCENTAGE Calculate commission
        by ((service_total_price*commission.figure_value)/100).
        """
        total_commission = 0.0
        service = self.service_details

        if not service.service_tier_commissions:
            return total_commission

        commission_obj = list(
            filter(
                lambda k: k.service_tier_fee_detail.id == service_tier.id,
                service.service_tier_commissions
            )
        )

        if not commission_obj:
            return total_commission

        commission_obj = commission_obj[0]

        if commission_obj.commission_type == "FLAT":
            total_commission = commission_obj.figure_value
        else:
            total_commission = self.percentage(service_price, commission_obj.figure_value)

        return total_commission


class CouponMixin(ServiceBase):
    """
    Coupon mixin class contains coupon related functions and validator.
    Add function and methods according to your need.
    """

    @classmethod
    def get_coupon_by_service(cls, service_id, coupon_code, to_dict=False):
        """
        Return elastic search dsl object of coupon or None.
        to_dict: True will return dict obj of coupon.
        """
        coupon_obj = None
        if service_id and coupon_code:
            result = Search(
                using=EsClient,
                index=SERVICE_CACHE_INDEX
            ).query('nested',
                    path='coupons',
                    query=Q('match',
                            coupons__code=coupon_code)).filter(
                Q('bool',
                  should=Q('match', id=service_id), minimum_should_match=1)).execute()
            if result and result.hits:
                coupon_obj = result.hits[0]
                if to_dict:
                    coupon_obj = coupon_obj.to_dict()

        return coupon_obj

    def get_coupon_from_service(self, coupon_code, service=None):
        """
        Returns coupon from service coupons list.
        """
        if not service:
            service = self.service_details

        coupon = list(
            filter(
                lambda coup: coup.code == coupon_code, service.coupons
            )
        )
        if coupon:
            return coupon[0]

        return None

    def get_coupon_details(self, coupon_code=None, **kwargs):
        """
        Fetch respective coupon from the service coupon index,
        Validate coupon price and apply on service total price.

        pass get_coupon_details(service_list=list(service_list))
        Service list is required to validate coupon of dependent services,
        If coupon is dependent on other service then we need to check that service
        is added in service list.
        """
        total_price = self.total_price
        coupons_to_apply = list()

        # service_list to validate dependent service condition
        service_list = kwargs.get('service_list') or list()
        coupon_code_list = kwargs.get('coupon_code_list') or list()
        service = self.service_details
        service_coupons = service.coupons or list()

        if coupon_code:
            coupon_code_list.append(coupon_code)

        for c_code in coupon_code_list:
            coupons_to_apply += list(
                filter(
                    lambda coup: coup.code == c_code, service_coupons
                )
            )

        for coupon in service_coupons:
            if coupon.is_auto_apply:
                coupons_to_apply.append(coupon)

        coupons_to_apply = list(
            filter(
                lambda coupon_obj: coupon_obj.min_amount_for_coupon < total_price, coupons_to_apply
            )
        )

        if coupons_to_apply:
            coupon_to_be_apply = self._get_coupon_with_max_discount(coupons_to_apply)
            if coupon_to_be_apply.dependent_services:
                is_valid_coupon = any(
                    dependent.service_id in service_list for dependent in coupon_to_be_apply.dependent_services
                )
                if not is_valid_coupon:
                    coupons_to_apply = self._remove_coupon_from_list(
                        coupon_code=coupon_to_be_apply.code,
                        coupon_list=coupons_to_apply
                    )
                    coupon_to_be_apply = self._get_coupon_with_max_discount(coupons_to_apply)

            if coupon_to_be_apply:
                self._total_discount = coupon_to_be_apply.max_discount
                self._coupon_code = coupon_to_be_apply.code

        return self._total_discount, self._coupon_code

    @classmethod
    def _get_coupon_with_max_discount(cls, coupon_list):
        if coupon_list:
            return max(coupon_list, key=lambda k: k.max_discount)

    @classmethod
    def _remove_coupon_from_list(cls, coupon_code, coupon_list):
        coupons_to_apply = []

        for coupon in coupon_list:
            if coupon.code != coupon_code:
                coupons_to_apply.append(coupon)

        return coupons_to_apply

    def validate_coupon_code(self, coupon_code: str, service_list=None, raise_exception=False, **kwargs):
        """
        Validates the coupon is valid with respect to service_list passed.
        Service list is required to validate coupon of dependent services,
        If coupon is dependent on other service then we need to check that service
        is added in service list.
        """
        is_valid_coupon = False
        service = self.service_details

        if not service_list:
            service_list = list()

        if service_list and kwargs.get('re_fetch'):
            coupon_search_query = Search(
                using=EsClient, index=SERVICE_CACHE_INDEX
            ).query('nested', path='coupons', query=Q('match', coupons__code=coupon_code)).filter(
                Q('bool',
                  should=[Q('match', id=service_id) for service_id in service_list],
                  minimum_should_match=1)
            )
            response = coupon_search_query.execute()
            if response and response.hits:
                is_valid_coupon = True
                service = response.hits[0]

        coupon = self.get_coupon_from_service(coupon_code, service)

        if coupon:
            if coupon.dependent_services:
                # Service list is required to validate coupon of dependent services,
                # If coupon is dependent on other service then we need to check that service
                # is added in service list.
                service_list = service_list or []
                is_valid_coupon = any(dependent.service_id in service_list for dependent in coupon.dependent_services)
            else:
                is_valid_coupon = True

        if not is_valid_coupon and raise_exception:
            raise APIException({'coupon_code': ['Invalid coupon code']})

        return is_valid_coupon


class ServicePrice(PriceCatalog, CouponMixin):
    """
    Service price class to fetch service fee structure from `Elasticsearch` index.
    service_id: Respective service id,
    units: Units/quantity of service
    -----------------------------------------------------------------------------
    Call `ServicePrice(service_id, units).get_fee_details()` to get service fee details.

    Add more functions & calculations acc. to your need.
    """

    pass


class ServiceValidator(object):
    """
    Validate service or service list passed to validator,
    search and validate service using service_id from `Elasticsearch`.
    """
    def __init__(self, service_id=None, raise_exception=False, **kwargs):
        """
        kwargs service_list: validates list of services.
        kwargs match_all: validates all services else validate at least one service.
        """
        self.is_valid_service = False
        self.service_list = kwargs.get('service_list') or list()
        self.minimum_should_match = 1
        if service_id:
            self.service_list.append(service_id)
        if kwargs.get('match_all'):
            self.minimum_should_match = len(self.service_list)

        self.is_valid_service = self.validate_service_list()

        if not self.is_valid_service and raise_exception:
            raise APIException("Invalid service")

    def validate_service_list(self):
        """
        Search and return bool of service exists in Elasticsearch or not.
        """
        result = Search(
            using=EsClient,
            index=SERVICE_CACHE_INDEX
        ).query(
            Q('bool',
              should=[Q('match', id=service_id) for service_id in self.service_list],
              minimum_should_match=self.minimum_should_match)
        ).execute()
        if result and result.hits:
            return True
        return False



import json
from flask import request
from .exceptions import logger


class CustomResponse(object):
    def __init__(self, response):
        self.response = response

    @property
    def render(self):
        message = None
        data = dict()
        status_code = self.response.status_code
        response_data = json.loads(self.response.data)
        if type(response_data) == dict:
            message = response_data.pop('message', None)
            data = response_data.pop('data', None) if response_data.get('data') else response_data

        if 400 <= status_code <= 499:
            message = response_data.copy() if not message else message
            response_data.clear()

        response_json = {
            "data": data,
            "message": message,
            "title": self.response.status,
            "code": self.response.status_code
        }
        if type(response_data) == dict:
            response_json.update(response_data)

        self.response.data = json.dumps(response_json)
        self.log(response_json)
        return self.response

    def log(self, data):
        if self.response.status_code == 400:
            log = '[{} {} {}]--{}'.format(
                self.response.status_code, request.method,
                request.full_path, data.get('message'))
            logger.error(log)



import jwt
from flask_babel import lazy_gettext as _
from .exceptions import AuthException
from .utils import get_user_obj


from settings import SIGNING_KEY, JWT_ALGORITHMS


def jwt_get_user_permissions_from_payload_handler(payload):
    """
    Override this function if permissions is formatted differently in payload
    """
    return payload.get('permissions')


def jwt_get_user_from_payload_handler(payload):
    """
    Override this function if permissions is formatted differently in payload
    """
    return payload.get('user', '') or None


def authenticate_credentials(payload):
    """
    Returns an active user that matches the payload's user id.
    """

    user = jwt_get_user_from_payload_handler(payload)
    # fetch the user data from the token.
    if not user:
        raise AuthException(_('Invalid Authorization header. No user found.'))

    permissions = jwt_get_user_permissions_from_payload_handler(payload)
    user = get_user_obj(user)

    user.permissions = permissions

    user.is_authenticated = True

    if user.user_type == 'ADMIN':
        user.is_superuser = True
    elif user.user_type == 'AGENT':
        user.is_agent = True

    return user


def get_authorization_header(header):
    """
    Return request's 'Authorization:' header.
    """
    auth = header.get('Authorization')
    return auth


def get_internal_key_header(header):
    """
    Return request's 'INTERNAL_SERVICE_KEY:' header.
    """
    internal_key = header.get('INTERNAL-SERVICE-KEY')
    return internal_key


def authenticate(token):
    """
    Returns a two-tuple of `User` and token if a valid signature has been
    supplied using JWT-based authentication.  Otherwise, returns `None`.
    """
    auth = token.split()
    auth_header_prefix = 'bearer'
    if auth[0].lower() != auth_header_prefix:
        raise AuthException(_('Invalid prefix.'))
    if len(auth) == 1:
        raise AuthException(
            _('Invalid Authorization header. No credentials provided.'))
    elif len(auth) > 2:
        raise AuthException(
            _('Invalid Authorization header or Credentials string or should not contain spaces.'))
    jwt_value = auth[1]
    try:
        payload = jwt.decode(jwt_value, SIGNING_KEY, algorithms=JWT_ALGORITHMS)
    except jwt.InvalidSignatureError:
        raise AuthException(_('Invalid signature.'))
    except jwt.ExpiredSignatureError:
        raise AuthException(_('Signature expired.'))
    except jwt.DecodeError:
        raise AuthException(_('Error decoding signature.'))
    except Exception:
        raise AuthException(_('Error decoding signature'))

    user = authenticate_credentials(payload)
    return user
